#include <FreeRTOS.h>
#include <task.h>
#include <queue.h>
#include <libopencm3/stm32/gpio.h>
#include <libopencm3/stm32/rcc.h>
#include <libopencm3/stm32/exti.h>
#include <libopencm3/cm3/nvic.h>

#ifdef __GNUC__
#define UNUSED(x) UNUSED_##x __attribute__((__unused__))
#else
#define UNUSED(x) UNUSED_##x
#endif

//add stack-overflow hook
extern void vApplicationStackOverflowHook(xTaskHandle *pxTask, signed portCHAR *pcTaskName);

// our tasks
void task_onboardled(void *args);
void task_buttonled(void *args);
static QueueHandle_t btn_q; // coda dall'ISR del pulsante a task_buttonled

/*
 * applicazione di esempio. controlla un bottone e 2 led
 * un led (quello built-in) semplicemente fa on/off (più on che off)
 *
 * un altro led (porta A8) si accende quando il bottone (porta A1) è premuto (mandato a VCC)
 * vengono usati thread, interrupt e code per controllarli
 *
 * vedi pag. 191 del libro
 *
 * vantaggi: più responsivo e senza avere roba che gira in busy loop
 *           message passing è concettualmente più pulito
 * svantaggi: le code occupano memoria
 *            gli interrupt sono abbastanza flessibili ma non infinitamente, per sfruttarli bene su tante
 *            porte bisogna incastrare bene le scelte dei pin
 *
 * MONTAGGIO
 * il pulsante va messo tra la porta A1 e il VCC
 * il led va messo tra la porta A8, una resistenza (330ohm) e la massa
 */

static void gpio_setup(void)
{

	// built-in LED
	rcc_periph_clock_enable(RCC_GPIOC);
	gpio_set_mode(GPIOC, GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL,
	              GPIO13);

	// ext. push button (yellow)
	rcc_periph_clock_enable(RCC_GPIOA);
	gpio_set_mode(GPIOA, GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, GPIO1);

    exti_select_source(EXTI1, GPIOA);
    exti_set_trigger(EXTI1, EXTI_TRIGGER_RISING);
    exti_enable_request(EXTI1);
    nvic_enable_irq(NVIC_EXTI1_IRQ); // occhio! guarda tabella 11-2 a pag. 209 o ti perdi su quale NVIC_EXTI usare


	// ext. led (red)
	gpio_set_mode(GPIOA, GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, GPIO5);
}


int main(void)
{
    rcc_clock_setup_in_hse_8mhz_out_72mhz();
	gpio_setup();
	gpio_set(GPIOA, GPIO5);

    btn_q = xQueueCreate(10, sizeof(char));
	xTaskCreate(task_buttonled, "BUTTONLED", 50, NULL, configMAX_PRIORITIES - 1,
	            NULL);
	xTaskCreate(task_onboardled, "ONBOARD", 50, NULL, configMAX_PRIORITIES - 1,
	            NULL);
	vTaskStartScheduler();
	for (;;) {
	}

	return 0;
}

void task_onboardled(void *UNUSED(args))
{
	while (1) {
		gpio_clear(GPIOC, GPIO13); /* LED on */
		vTaskDelay(pdMS_TO_TICKS(500));

		gpio_set(GPIOC, GPIO13); /* LED off */
		vTaskDelay(pdMS_TO_TICKS(100));
	}
}

void task_buttonled(void *UNUSED(args))
{
    int ledOn = 1;
    while (1) {
        char value;
        if(xQueueReceive(btn_q, &value, pdMS_TO_TICKS(5000)) == pdPASS)
        {
            ledOn = !ledOn;
            if(ledOn != 0) {
                gpio_set(GPIOA, GPIO5);
            } else {
                gpio_clear(GPIOA, GPIO5);
            }
        }
    }
}

// gestisce l'interrupt del bottone. Viene chiamato quando il pulsante viene premuto, non quando viene
// rilasciato
void exti1_isr(void)
{
    xQueueSend(btn_q, (char) 0, 0);
    exti_reset_request(EXTI1);
    return;
}

// three "small flashes" (an S in Morse code) on onboard led
//
// questa funzione viene chiamata in caso di stackoverflow su uno dei task; e'
// quindi utile metterci un "segnale" riconoscibile per debug. In questo caso 3
// flash rapidi.
//
// siccome "tarare" la dimensione dello stack necessario e' difficile, con
// questa funzione si puo' andare "a tentoni" e ridurre la dimensione finche'
// non si ottengono crash
//
// in questa funzione, a quanto pare (ma non ho trovato doc) non si possono
// usare cose come vTaskDelay. Siamo quindi privi di RTOS e per fare gli sleep
// dobbiamo fare i for "all'antica"

void vApplicationStackOverflowHook(xTaskHandle *UNUSED(pxTask),
        signed portCHAR *UNUSED(pcTaskName))
{
    int i;
    for (;;) {
        for (char c = 0; c < 3; c++) {
            gpio_clear(GPIOC, GPIO13); /* LED on */
            for (i = 0; i < 1000000;
                    i++)                       /* small delay */
                __asm__("nop");

            gpio_set(GPIOC, GPIO13);       /* LED off */
            for (i = 0; i < 3000000; i++) { /* a bit more delay */
                __asm__("nop");
            }
        }
        for (i = 0; i < 10000000; i++) { /* long delay */
            __asm__("nop");
        }
    }
}
// vim: set fdm=marker:
