#include <FreeRTOS.h>
#include <task.h>
#include <portable.h>
#include <queue.h>
#include <string.h>

#include <libopencm3/stm32/gpio.h>
#include <libopencm3/stm32/rcc.h>
#include <libopencm3/stm32/exti.h>
#include <libopencm3/stm32/usart.h>
#include <libopencm3/cm3/nvic.h>

#include <lib/mini-printf.h>

#ifdef __GNUC__
#define UNUSED(x) UNUSED_##x __attribute__((__unused__))
#else
#define UNUSED(x) UNUSED_##x
#endif

//add stack-overflow hook
extern void vApplicationStackOverflowHook(xTaskHandle *pxTask, signed portCHAR *pcTaskName);

// our tasks
void task_onboardled(void *args);
void task_buttonled(void *args);
void task_uart_tx(void *args);
void task_uart_rx(void *args);
static QueueHandle_t btn_q; // coda dall'ISR del pulsante a task_buttonled
static QueueHandle_t uart_tx_q;

/*
 * Partiamo dall'esempio `rotary` e aggiungiamo comunicazione seriale. Vediamo come fare un sistema in cui si
 * possa sia "rispondere" a comandi del computer sia "iniziare" noi una comunicazione.
 * In particolare il computer ci puo' chiedere il duty cycle attuale (vedi task_uart_rx); di contro noi lo
 * aggiorniamo quando questo cambia.
 *
 * per interagire dal computer a me picocom piace piu' di minicom perché puoi dire tutto da commandline; si
 * esce con C-a C-x
 *     picocom   --echo  /dev/ttyUSB0 --baud 38400  --databits 8 --stopbits 1 --parity n --imap lfcrlf
 *
 * Siccome ora abbiamo una seriale ha senso lavorare con le stringhe. NON usiamo la sprintf "normale" perche'
 * occupa tantissimo in memoria e blocca tutto! Usiamo una implementazione minimale della snprintf, la
 * mini-printf. Questo cambiamento va riportato nel Makefile!
 * Sempre siccome abbiamo le stringhe e vogliamo metterle in una coda, dobbiamo vedere la malloc, o meglio la
 * pvPortMalloc. Per usarla dobbiamo includere portable.h. Serve anche mettere rtos/heap_4.c nel Makefile,  ma
 * c'era gia'.
 * 
 */

static void gpio_setup(void)
{

	// built-in LED
	rcc_periph_clock_enable(RCC_GPIOC);
	gpio_set_mode(GPIOC, GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL,
	              GPIO13);

	// ext. rotary button: due pin per la direzione (A3 e A4). Metto l'interrupt solo su uno dei due (B1)
	rcc_periph_clock_enable(RCC_GPIOA);
	gpio_set_mode(GPIOA, GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, GPIO0|GPIO1);

    exti_select_source(EXTI0, GPIOA);
    exti_set_trigger(EXTI0, EXTI_TRIGGER_RISING);
    exti_enable_request(EXTI0);
    nvic_enable_irq(NVIC_EXTI0_IRQ); // occhio! guarda tabella 11-2 a pag. 209 o ti perdi su quale NVIC_EXTI usare

	// ext. led (red)
	gpio_set_mode(GPIOA, GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, GPIO8);
	// ext. led (green)
	rcc_periph_clock_enable(RCC_GPIOB);
	gpio_set_mode(GPIOB, GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, GPIO11);

    // Serial comunication
    rcc_periph_clock_enable(RCC_USART1);
    gpio_set_mode(GPIOA, GPIO_MODE_OUTPUT_50_MHZ,
            GPIO_CNF_OUTPUT_ALTFN_PUSHPULL,
            GPIO_USART1_TX);
    usart_set_baudrate(USART1, 38400);
    usart_set_databits(USART1,8);
    usart_set_stopbits(USART1,USART_STOPBITS_1);
    usart_set_mode(USART1,USART_MODE_TX_RX);
    usart_set_parity(USART1,USART_PARITY_NONE);
    usart_set_flow_control(USART1,USART_FLOWCONTROL_NONE);
    usart_enable(USART1);
}

static void uart_puts(const char *s) {
    for(; *s; ++s) {
        usart_send_blocking(USART1, *s);
    }
}


int main(void)
{
    rcc_clock_setup_in_hse_8mhz_out_72mhz();
	gpio_setup();
    gpio_set(GPIOA, GPIO8);
    gpio_clear(GPIOB, GPIO11);
    uart_puts("READY\n");

    btn_q = xQueueCreate(10, sizeof(char));
    uart_tx_q = xQueueCreate(10, sizeof(char*));
	xTaskCreate(task_buttonled, "BUTTONLED", 300, NULL, configMAX_PRIORITIES - 1,
	            NULL);
	xTaskCreate(task_onboardled, "ONBOARD", 50, NULL, configMAX_PRIORITIES - 1,
	            NULL);
	xTaskCreate(task_uart_tx, "TX", 50, NULL, configMAX_PRIORITIES - 1,
	            NULL);
	xTaskCreate(task_uart_rx, "RX", 50, NULL, configMAX_PRIORITIES - 1,
	            NULL);
	vTaskStartScheduler();
	for (;;) {
	}

	return 0;
}

int boardLedDuty = 1000;
void task_onboardled(void *UNUSED(args))
{
	while (1) {
		gpio_clear(GPIOC, GPIO13); /* LED on */
		vTaskDelay(pdMS_TO_TICKS(boardLedDuty));

		gpio_set(GPIOC, GPIO13); /* LED off */
		vTaskDelay(pdMS_TO_TICKS(500));
	}
}

void task_buttonled(void *UNUSED(args))
{
    while (1) {
        char increasing;
        if(xQueueReceive(btn_q, (void*)&increasing, pdMS_TO_TICKS(5000)) == pdPASS)
        {
            if(increasing) {
                gpio_toggle(GPIOA, GPIO8);
            } else {
                gpio_toggle(GPIOB, GPIO11);
            }
            char *buf = (void*) pvPortMalloc(16);
            if(buf != NULL) {
                snprintf(buf, 16, "%c %d\n", increasing ? 'U' : 'D', boardLedDuty);
                // attenzione: xQueueSend vuole un puntatore alle cose che deve copiare.
                // noi dobbiamo mandare una stringa, quindi un puntatore. Dobbiamo quindi mandare un puntatore
                // a puntatore!
                if(xQueueSend(uart_tx_q, (void*) &buf, 0) != pdTRUE) {
                    // la free la fa il task che legge; ma se non riusciamo a mettere in coda (coda piena) 
                    // allora la free dobbiamo farla noi
                    vPortFree(buf);
                }
            }
        }
    }
}

void task_uart_tx(void *UNUSED(args)) {
    while (1) {
        char *buf, *buf2;
        if(xQueueReceive(uart_tx_q, (void*)&buf, pdMS_TO_TICKS(5000)) != pdPASS)
            continue;
        buf2 = buf; // mi devo tenere il puntatore originale senno' dopo non posso fare free
        for(; *buf != '\0'; buf++) {
            while(!usart_get_flag(USART1, USART_SR_TXE))
                taskYIELD();
            usart_send_blocking(USART1, *buf);
        }
        vPortFree(buf2);
    }
}

void task_uart_rx(void *UNUSED(args)) {
    char cmd[32];
    unsigned char i = 0;
    while (1) {
        if (i >= 30) { // avoid overflows
            i = 0;
        }
        cmd[i] = usart_recv_blocking(USART1);
        if (cmd[i] == '\r' || cmd[i] == '\n') {
            gpio_toggle(GPIOB, GPIO11);
            cmd[i] = '\0'; // replace newline with end-of-string
            if (strcmp(cmd, "GET") == 0) { //check valid commands
                char *buf = (void *)pvPortMalloc(16);
                if (buf != NULL) {
                    snprintf(buf, 16, "-> %d\n", boardLedDuty);
                    // attenzione: xQueueSend vuole un puntatore alle cose che deve
                    // copiare. noi dobbiamo mandare una stringa, quindi un puntatore.
                    // Dobbiamo quindi mandare un puntatore a puntatore!
                    if (xQueueSend(uart_tx_q, (void *)&buf, 0) != pdTRUE) {
                        // la free la fa il task che legge; ma se non riusciamo a mettere in
                        // coda (coda piena) allora la free dobbiamo farla noi
                        vPortFree(buf);
                    }
                }
            } // GET
            i=0;
        } // newline
        else {
            i++;
        }
    }
}

// gestisce l'interrupt del bottone. Viene chiamato quando il pulsante viene premuto, non quando viene
// rilasciato
void exti0_isr(void)
{
    char v = gpio_get(GPIOA, GPIO1);
    exti_reset_request(EXTI0);
    if(v) { // decrease
        if(boardLedDuty < 220)
            boardLedDuty = 30;
        else
            boardLedDuty -= 200;
        xQueueSend(btn_q, (void*) 0, 0);
    } else {
        if(boardLedDuty >= 2300)
            boardLedDuty = 2500;
        else
            boardLedDuty += 200;
        xQueueSend(btn_q, (void*) 1, 0);
    }
}

// three "small flashes" (an S in Morse code) on onboard led
//
// questa funzione viene chiamata in caso di stackoverflow su uno dei task; e'
// quindi utile metterci un "segnale" riconoscibile per debug. In questo caso 3
// flash rapidi.
//
// siccome "tarare" la dimensione dello stack necessario e' difficile, con
// questa funzione si puo' andare "a tentoni" e ridurre la dimensione finche'
// non si ottengono crash
//
// in questa funzione, a quanto pare (ma non ho trovato doc) non si possono
// usare cose come vTaskDelay. Siamo quindi privi di RTOS e per fare gli sleep
// dobbiamo fare i for "all'antica"

void vApplicationStackOverflowHook(xTaskHandle *UNUSED(pxTask),
        signed portCHAR *UNUSED(pcTaskName))
{
    int i;
    for (;;) {
        for (char c = 0; c < 3; c++) {
            gpio_clear(GPIOC, GPIO13); /* LED on */
            for (i = 0; i < 1000000;
                    i++)                       /* small delay */
                __asm__("nop");

            gpio_set(GPIOC, GPIO13);       /* LED off */
            for (i = 0; i < 3000000; i++) { /* a bit more delay */
                __asm__("nop");
            }
        }
        for (i = 0; i < 10000000; i++) { /* long delay */
            __asm__("nop");
        }
    }
}

// vim: set fdm=marker:
